package com.jdbernard.wdiwtlt.db

import java.lang.reflect.Modifier
import java.sql.Connection
import java.sql.PreparedStatement
import java.sql.ResultSet
import java.sql.Timestamp
import java.text.SimpleDateFormat
import java.util.regex.Pattern
import javax.persistence.Entity
import javax.sql.DataSource

import groovy.sql.Sql
import groovy.transform.CompileStatic
import groovy.transform.TailRecursive

import org.slf4j.Logger
import org.slf4j.LoggerFactory

import com.jdbernard.wdiwtlt.db.models.*

public class DbApi {

    private DataSource dataSource
    private Sql sql

    private static UPPERCASE_PATTERN = Pattern.compile(/(.)(\p{javaUpperCase})/)
    private static Logger logger = LoggerFactory.getLogger(DbApi)

    public DbApi(DataSource dataSource) {
        this.dataSource = dataSource
        this.sql = new Sql(dataSource) }

    public void shutdown() { dataSource.shutdown() }

    /// ### Common
    public def getAll(Class modelClass, String orderClause = null) {
        def query = new StringBuilder()
        query.append('SELECT * FROM ')
             .append(pluralize(nameFromModel(modelClass.simpleName)))

        if (orderClause) query.append(' ORDER BY ').append(orderClause)

        query = query.toString()

        logger.debug('Selecting models.\n\tSQL: {}', query)
        return sql.rows(query).collect { recordToModel(modelClass, it) } }

    public def getById(Class modelClass, int id) {
        def model = modelClass.newInstance()
        model.id = id
        return refresh(model) }

    public def refresh(def model) {
        def query = new StringBuilder()
            .append('SELECT * FROM ')
            .append(pluralize(nameFromModel(model.class.simpleName)))
            .append(' WHERE id = ?')
            .toString()

        logger.debug('Selecting model.\n\tSQL: {}\n\tPARAMS: {}', query, model.id)
        def result = sql.firstRow(query, [model.id])
        if (!result) return null
        return updateModel(result, model) }

    public def getByName(Class modelClass, String name) {
        def query = new StringBuilder()
            .append('SELECT * FROM ')
            .append(pluralize(nameFromModel(modelClass.simpleName)))
            .append(' WHERE name = ?')
            .toString()

        logger.debug('Selecting model.\n\tSQL: {}\n\tPARAMS: {}', query, name)
        return sql.rows(query, [name]).collect { recordToModel(modelClass, it) } }

    public def getBy(Class modelClass, List<String> columns,
    List<Object> values) {
        def query = new StringBuilder()
            .append('SELECT * FROM ')
            .append(pluralize(nameFromModel(modelClass.simpleName)))
            .append(' WHERE ')
            .append(columns.collect { it + ' = ?' }.join(' AND '))
            .toString()

        logger.debug('Selecting models.\n\tSQL: {}\n\tPARAMS: {}', query, values)
        return sql.rows(query, values)
            .collect { recordToModel(modelClass, it) } }

    public def getLike(Class modelClass, List<String> columns,
    List<Object> values) {
        values = values.collect { "%$it%".toString() }
        String query = new StringBuilder()
            .append('SELECT * FROM ')
            .append(pluralize(nameFromModel(modelClass.simpleName)))
            .append(' WHERE ')
            .append(columns.collect { """"$it" LIKE ?"""}.join(' AND '))
            .toString()

        logger.debug('Selecting models.\n\tSQL: {}\n\tPARAMS: {}', query, values)
        return sql.rows(query, values)
            .collect { recordToModel(modelClass, it) } }

    public def getWhere(Class modelClass, Map criteria) {
        switch(modelClass) {
            case Album:      return getAlbumsWhere(criteria)
            case Artist:     return getArtistsWhere(criteria)
            case Bookmark:   return getBookmarksWhere(criteria)
            case MediaFile:  return getMediaFilesWhere(criteria)
            case Playlist:   return getPlaylistsWhere(criteria)
            case Tag:        return getTagsWhere(criteria) } }

    public def save(def model) {
        if (model.id > 0) return update(model)
        else return create(model) }

    public def update(def model) {
        def setClauses = []
        def params = []

        getInstanceFields(model.class)
            .findAll { it.name != 'id' }
            .each { field ->
                setClauses << '"' + nameFromModel(field.name) + '"= ?'
                params << field.get(model) }

        def query = new StringBuilder()
            .append('UPDATE ')
            .append(pluralize(nameFromModel(model.class.simpleName)))
            .append(' SET ')
            .append(setClauses.join(', '))
            .append(' WHERE id = ?')
            .toString()

        params << model.id

        return withTransaction {
            logger.debug('Updating model.\n\tSQL: {}\n\tPARAMS: {}', query, params)
            sql.executeUpdate(query, params)
            return refresh(model) } }

    public def create(def model) {
        def columns = []
        def params = []

        getInstanceFields(model.class)
            .findAll { it.name != 'id' }
            .each { field ->
                //if (field.class.getAnnotation(Entity)) // check to see if we
                                                        // have nested models
                columns << '"' + nameFromModel(field.name) + '"'
                params << field.get(model) }

        def query= new StringBuilder()
            .append('INSERT INTO ')
            .append(pluralize(nameFromModel(model.class.simpleName)))
            .append(' (')
            .append(columns.join(', '))
            .append(') VALUES (')
            .append((1..columns.size()).collect { '?' }.join(', '))
            .append(')').toString()

        logger.debug('Creating model.\n\tSQL: {}\n\tPARAMS: {}', query, params)
        model.id = sql.executeInsert(query, params)[0][0]
        return model }

    public def delete(def model) {
        def query = new StringBuilder()
            .append('DELETE FROM ')
            .append(pluralize(nameFromModel(model.class.simpleName)))
            .append('WHERE id = ?')
            .toString()

        logger.debug('Deleting model.\n\tSQL: {}\n\tPARAMS: {}',
            query, model.id)
        sql.execute(query, [model.id])
        return sql.updateCount }

    public def associate(Class modelClass1, Class modelClass2 ,
    Integer firstId, Integer secondId) {
        String linkTable = pluralize(nameFromModel(modelClass1.simpleName)) +
                     '_' + pluralize(nameFromModel(modelClass2.simpleName))
        String col1 = nameFromModel(modelClass1.simpleName) + '_id'
        String col2 = nameFromModel(modelClass2.simpleName) + '_id'

        return withTransaction {
            def query = """\
                SELECT * FROM $linkTable
                WHERE "${col1}" = ? AND "${col2}" = ?"""
            def params = [firstId, secondId]

            // Look first for an existing association before creating one.
            logger.debug('Selecting association.\n\tSQL: {}\n\tPARAMS: {}',
                query, params)
            if (sql.firstRow(query, params)) { return 0 }
            else {
                query = """INSERT INTO $linkTable ("$col1", "$col2") VALUES (?, ?)"""
                logger.debug('Creating association.\n\tSQL: {}\n\tPARAMS: {}',
                    query, params)
                return sql.execute(query, params) } } }

    public def associate(def m1, def m2) {
        return associate(m1.class, m2.class, m1.id, m2.id) }

    /// ### Album-specific methods
    //  =======================================================================
    public Album getAlbumById(int id) { return getById(Album, id) }

    public List<Album> getAlbums() { return getAll(Album) }

    public List<Album> getAlbumsOrderedByName() {
        return getAll(Album, 'name ASC') }

    public List<Album> getAlbumsLikeName(String name) {
        return getLike(Album, ['name'], [name]) }

    public List<Album> getAlbumsWhere(Map params) {
        def query = new StringBuilder()
        def sqlParams = []

        query.append('SELECT DISTINCT al.* FROM albums al ')

        if (params.artistId) {
            query.append(' JOIN artists_albums aa ON al.id = aa.album_id ')
            query.append(' AND aa.artist_id = ? ')
            sqlParams << params.artistId }

        if (params.mediaFileId || params.playlistId) {
            query.append(' JOIN albums_media_files amf ON ')
            query.append(' al.id = amf.album_id ')

            if (params.mediaFileId) {
                query.append(' AND amf.media_file_id = ? ')
                sqlParams << params.mediaFileId } }

        if (params.playlistId) {
            query.append(' JOIN playlists_media_files pmf ON ')
            query.append(' amf.media_file_id = pmf.media_file_id AND ')
            query.append(' pmf.playlist_id = ?')
            sqlParams << params.playlistId }

        if (params.name || params.year) { query.append(' WHERE ') }

        if (params.year) {
            query.append(' al.year = ? ')
            sqlParams << params.year }

        if (params.name) {
            if (params.year) query.append(' AND ')
            query.append(' al.name = ? ')
            sqlParams << params.name }

        query = query.toString()
        logger.debug('Selecting albums.\n\tSQL: {}\n\tPARAMS: {}',
            query, sqlParams)
        return sql.rows(query, sqlParams).collect { recordToModel(Album, it) } }

    public void removeEmptyAlbums() {

        String query = """\
            DELETE FROM albums WHERE id IN (
                SELECT DISTINCT al.id
                FROM albums al LEFT OUTER JOIN albums_media_files almf ON
                    al.id = almf.album_id
                WHERE almf.album_id IS NULL)"""

        logger.debug('Deleting empty albums.\n\tSQL: {}', query)
        sql.execute(query) }

    /// ### Artist-specific methods
    public List<Artist> getArtists() { return getAll(Artist) }
    public Artist getArtistById(int id) { return getById(Artist, id) }

    public Artist getArtistByName(String name) {
        def artists = getByName(Artist, name)
        if (artists) return artists[0] }

    public List<Artist> getArtistsOrderedByName() {
        return getAll(Artist, 'name ASC') }

    public List<Artist> getArtistsWhere(Map params) {
        def query = new StringBuilder()
        def sqlParams = []

        query.append('SELECT DISTINCT ar.* FROM artists ar ')

        if (params.albumId) {
            query.append(' JOIN artists_albums aa ON ar.id = aa.artist_id ')
            query.append(' AND aa.album_id = ? ')
            sqlParams << params.albumId }

        if (params.mediaFileId || params.playlistId) {
            query.append(' JOIN artists_media_files amf ON ')
            query.append(' ar.id = amf.artist_id ')

            if (params.mediaFileId) {
                query.append(' AND amf.media_file_id = ? ')
                sqlParams << params.mediaFileId } }

        if (params.playlistId) {
            query.append(' JOIN playlists_media_files pmf ON ')
            query.append(' amf.media_file_id = pmf.media_file_id AND ')
            query.append(' pmf.playlist_id = ?')
            sqlParams << params.playlistId }

        if (params.name) {
            query.append(' WHERE ar.name = ? ')
            sqlParams << params.name }

        query = query.toString()
        logger.debug('Selecting artists.\n\tSQL: {}\n\tPARAMS: {}',
            query, sqlParams)
        return sql.rows(query, sqlParams).collect { recordToModel(Artist, it) } }

    public List<Artist> getArtistsLikeName(String name) {
        return getLike(Artist, ['name'], [name]) }

    public def addAlbumArtist(int albumId, int artistId) {
        return associate(Artist, Album, artistId, albumId) }

    public void removeEmptyArtists() {

        String query = """\
            DELETE FROM artists WHERE id IN (
                SELECT DISTINCT ar.id
                FROM artists ar LEFT OUTER JOIN artists_media_files armf ON
                    ar.id = armf.artist_id
                WHERE armf.artist_id IS NULL)"""

        logger.debug('Deleting empty artists.\n\tSQL: {}', query)
        sql.execute(query) }

    /// ### Bookmark-specific methods
    public Bookmark getBookmarkById(int id) { return getById(Bookmark, id) }

    public List<Bookmark> getBookmarkByName(String name) {
        return getByName(Bookmark, name) }

    public List<Bookmark> getBookmarks() { return getAll(Bookmark) }
    public List<Bookmark> getBookmarksOrderedByName() {
        return getAll(Bookmark, 'name ASC') }

    public List<Bookmark> getBookmarksWhere(Map params) {
        def query = new StringBuilder()
        def sqlParams = []
        def orderClauses = []
        def whereClauses = []

        query.append('SELECT b.* FROM bookmarks b ')

        if (params.playlistId) {
            whereClauses << 'b.playlist_id = ?'
            sqlParams << params.playlistId }

        if (params.userCreated != null) {
            whereClauses << 'b.user_created = ?'
            sqlParams << params.userCreated }

        if (params.name) {
            whereClauses << 'b.name = ?'
            sqlParams << params.name }

        if (params.mediaFileId) {
            whereClauses << 'b.media_file_id = ?'
            sqlParams << params.mediaFileId }

        if (params.playIndex) {
            whereClauses << 'b.play_index = ?'
            sqlParams << params.playIndex }

        if (params.lastUsedBefore) {
            whereClauses << 'b.last_used < ?'
            sqlParams << params.lastUsedBefore }

        if (params.lastUsedAfter) {
            whereClauses << 'b.last_used > ?'
            sqlParams << params.lastUsedAfter }

        if (params.lastUsedBetween) {
            whereClauses << 'b.last_used BETWEEN ? AND ?'
            sqlParams.addAll(params.lastUsedBetween) }

        orderClauses << 'b.last_used DESC'

        if (whereClauses)
            query.append(' WHERE ').append(whereClauses.join(' AND '))

        if (orderClauses)
            query.append(' ORDER BY ').append(orderClauses.join(', '))

        query = query.toString()
        logger.debug('Selecting bookmarks.\n\tSQL: {}\n\tPARAMS: {}',
            query, sqlParams)
        return sql.rows(query, sqlParams)
            .collect { recordToModel(Bookmark, it) } }

    /// ### Image-specific methods
    public Image getImageById(int id) { return getById(Image, id) }

    /// ### MediaFile-specific methods
    public MediaFile getMediaFileById(int id) { return getById(MediaFile, id) }

    public List<MediaFile> getMediaFileByName(String name) {
        return getByName(MediaFile, name) }

    public List<MediaFile> getMediaFiles() { return getAll(MediaFile) }

    public List<MediaFile> getMediaFilesLikeName(String name) {
        return getLike(MediaFile, ['name'], [name]) }

    public MediaFile getMediaFileByFilePath(String filePath) {
        def files = getBy(MediaFile, ['file_path'], [filePath])
        return files ? files[0] : null }

    public MediaFile getMediaFileByFileHash(String fileHash) {
      def files = getBy(MediaFile, ['file_hash'], [fileHash])
      return files ? files[0] : null }

    public List<MediaFile> getMediaFilesWhere(Map params) {
        def query = new StringBuilder()
        def sqlParams = []
        def joins = []
        def orderClauses = [] as LinkedList
        def whereClauses = []

        def newJoin
        query.append('SELECT mf.* FROM media_files mf ')

        if (params.artistId || params.playlistId) {
            joins << (newJoin = [
                table: 'artists_media_files armf',
                clauses: ['mf.id = armf.media_file_id']])

            if (params.artistId) {
                newJoin.clauses << 'armf.artist_id = ?'
                sqlParams << params.artistId }
            else newJoin.type = 'LEFT'

            orderClauses << 'armf.artist_id ASC' }

        if (params.albumId || params.artistId || params.playlistId) {
            joins << (newJoin = [
                table: 'albums_media_files almf',
                clauses: ['mf.id = almf.media_file_id']])

            if (params.albumId) {
                newJoin.clauses << 'almf.album_id = ?'
                sqlParams << params.albumId }
            else newJoin.type = 'LEFT'
            orderClauses << 'almf.album_id ASC' }

        if (params.playlistId) {
            joins << (newJoin = [
                table: 'playlists_media_files pmf',
                clauses: ['pmf.media_file_id = mf.id', 'pmf.playlist_id = ?']])
            orderClauses.addFirst('pmf.position')
            sqlParams << params.playlistId }

        if (params.name) {
            whereClauses << 'mf.name = ?'
            sqlParams << params.name }

        if (params.discNumber) {
            query.append('mf.disc_number = ?')
            sqlParams << params.discNumber }

        if (params.trackNumber) {
            query.append('mf.track_number = ?')
            sqlParams << params.trackNumber }

        if (params.filePath) {
            query.append('mf.file_path = ?')
            sqlParams << params.filePath }

        if (params.fileHash) {
            query.append('mf.file_hash = ?')
            sqlParams << params.fileHash }

        if (params.tags) {
            params.tags.eachWithIndex { tag, idx ->
                String L = "mft${idx}";
                joins << (newJoin = [
                    table: "media_files_tags $L",
                    clauses: ["${L}.media_file_id = mf.id",
                              "${L}.tag_id = ?"]])
                sqlParams << tag.id } }

        orderClauses << 'mf.disc_number'
        orderClauses << 'mf.track_number'

        joins.each { j ->
            if (j.type) query.append(' ').append(j.type).append(' ')

            query.append(' JOIN ')
                 .append(j.table)
                 .append(' ON ')
                 .append(j.clauses.join(' AND ')) }

        if (whereClauses)
            query.append(' WHERE ').append(whereClauses.join(' AND '))

        if (orderClauses)
            query.append(' ORDER BY ').append(orderClauses.join(', '))

        query = query.toString()
        logger.debug('Selecting media files.\n\tSQL: {}\n\tPARAMS: {}',
            query, sqlParams)
        return sql.rows(query, sqlParams).collect {
            recordToModel(MediaFile, it) }.unique() }

    public List<MediaFile> getUntaggedFiles() {
        String query = """
            SELECT mf.*
            FROM media_files mf LEFT JOIN
                 media_files_tags mft ON
                    mf.id = mft.media_file_id
            WHERE mft.media_file_id IS NULL"""

        logger.debug('Finding untagged media files.\n\tSQL: {}', query)
        return sql.rows(query).collect { recordToModel(MediaFile, it) } }

    public def associateMediaFileWithAlbum(int mediaFileId, int albumId) {
        return associate(Album, MediaFile, albumId, mediaFileId) }

    public def associateMediaFileWithArtist(int mediaFileId, int artistId) {
        return associate(Album, MediaFile, artistId, mediaFileId) }

    public def incrementPlayCount(int mediaFileId) {
        def query = 'UPDATE media_files SET play_count = play_count + 1 WHERE ID = ?'
        def params = [mediaFileId]

        logger.debug('Updating media file.\n\tSQL: {}\n\tPARAMS: {}', query, params)
        sql.executeUpdate(query, params)

        query = 'SELECT play_count FROM media_files WHERE id = ?'
        logger.debug('Selecting media file play count.\n\tSQL: {}\n\tPARAMS: {}', query, params)
        return sql.firstRow(query, params)[0] }

    public MediaFile incrementPlayCount(MediaFile mf) {
        mf.playCount = incrementPlayCount(mf.id)
        return mf }

    /// ### Playlist-specific methods
    public List<Playlist> getPlaylists() { return getAll(Playlist) }
    public Playlist getPlaylistById(int id) { return getById(Playlist, id) }

    public List<Playlist> getPlaylistByName(String name) {
        return getByName(Playlist, name) }

    public List<Playlist> getPlaylistsWhere(Map params) {
        def query = new StringBuilder()
        def sqlParams = []
        def orderClauses = []
        def whereClauses = []

        query.append('SELECT DISTINCT p.* FROM playlists p ')

        if (params.albumId || params.artistId || params.mediaFileId) {
            query.append(' JOIN playlists_media_files pmf ON ')
                 .append(' pmf.media_file_id = mf.id ') }

        if (params.mediaFileId) {
            query.append(' AND mf.media_file_id = ? ')
            sqlParam << params.mediaFileId }

        if (params.albumId) {
            query.append(' JOIN albums_media_files almf ON ')
                 .append(' mf.album_id = almf.album_id AND almf.album_id = ? ')
            sqlParams << params.albumId }

        if (params.artistId) {
            query.append(' JOIN arists_media_files armf ON ')
                 .append(' mf.artist_id = armf.artist_id AND ')
                 .append(' armf.artist_id = ? ')
            sqlParams << params.artistId }

        if (params.userCreated != null) {
            whereClauses << 'p.user_created = ?'
            sqlParams << params.userCreated }

        if (params.name) {
            whereClauses << 'p.name = ?'
            sqlParams << params.name }

        if (params.copiedFromId) {
            whereClauses << 'p.copied_from_id = ?'
            sqlParams << params.copiedFromId }

        if (params.lastUsedBefore) {
            whereClauses << 'p.last_used < ?'
            sqlParams << params.lastUsedBefore }

        if (params.lastUsedAfter) {
            whereClauses << 'p.last_used > ?'
            sqlParams << params.lastUsedAfter }

        if (params.lastUsedBetween) {
            whereClauses << 'p.last_used BETWEEN ? AND ?'
            sqlParams.addAll(params.lastUsedBetween) }

        orderClauses << 'p.last_used DESC'

        if (whereClauses)
            query.append(' WHERE ').append(whereClauses.join(' AND '))

        if (orderClauses)
            query.append(' ORDER BY ').append(orderClauses.join(', '))

        query = query.toString()
        logger.debug('Selecting playlists.\n\tSQL: {}\n\tPARAMS: {}',
            query, sqlParams)
        return sql.rows(query, sqlParams)
            .collect { recordToModel(Playlist, it) } }

    public int getNextPlaylistPosition(int playlistId) {
        String query = """\
            SELECT COALESCE(MAX(position), 0) + 1
            FROM playlists_media_files
            WHERE playlist_id = ?"""

        logger.debug('Finding next playlist position.\n\tSQL: {}\n\tPARAMS: {}',
            query, playlistId)

        return sql.firstRow(query, playlistId)[0] }

    private int incrementPlaylistPositions(int playlistId, int startPosition,
    int incrementAmount) {
        String query =  """\
            UPDATE playlists_media_files
            SET position = position + ?
            WHERE playlist_id = ? AND position > ?"""
        def params = [incrementAmount, playlistId, startPosition]

        logger.debug('Incrementing playlist positions.\n\tSQL: {}\n\tPARAMS: {}',
            query, params)
        return sql.executeUpdate(query, params) }

    public Playlist addToPlaylist(int playlistId,
    List<Integer> mediaFileIds, int startPosition = -1) {

        String query
        def params

        return withTransaction {
            def p = getById(Playlist, playlistId)
            if (!p) return null

            if (startPosition < 0)
                startPosition = getNextPlaylistPosition(playlistId)
            else
                incrementPlaylistPositions(playlistId, startPosition,
                    mediaFileIds.size())

            def position = startPosition
            mediaFileIds.each { mediaFileId ->
                query = 'INSERT INTO playlists_media_files VALUES (?, ?, ?)'
                params = [playlistId, mediaFileId, position++]
                logger.debug(
                    'Adding a new playlist/media file relation.\n\tSQL: {}\n\tPARAMS: {}',
                    query, params)
                sql.executeInsert(query, params) }

            p.mediaFileCount += mediaFileIds.size()
            p.lastUsed = new Timestamp(new Date().time)
            return update(p) } }

    public Playlist addToPlaylist(int playlistId, int mediaFileId,
    int position = -1) {

        String query
        def params

        return withTransaction {
            def p = getById(Playlist, playlistId)
            if (!p) return null

            if (position < 0) position = getNextPlaylistPosition(playlistId)
            else incrementPlaylistPositions(playlistId, position, 1)

            query = 'INSERT INTO playlists_media_files VALUES (?, ?, ?)'
            params = [playlistId, mediaFileId, position]

            logger.debug(
                'Adding a new playlist/media file relation.\n\tSQL: {}\n\tPARAMS: {}',
                query, params)
            sql.executeInsert(query, params)

            p.mediaFileCount++
            p.lastUsed = new Timestamp(new Date().time)
            return update(p) } }

    public Playlist removeFromPlaylist(int playlistId, int mediaFileId) {
        String getPositionQuery = """\
            SELECT position FROM playlists_media_files
            WHERE playlist_id = ? AND media_file_id = ?"""

        String delQuery = """\
            DELETE FROM playlists_media_files
            WHERE playlist_id = ? AND media_file_id = ?"""

        String reorderPlaylistQuery = """\
            UPDATE playlists_media_files SET position = position - 1
            WHERE playlist_id = ? AND position > ?"""

        String updatePlaylistQuery = """\
            UPDATE playlists SET
                media_file_count = (SELECT count(*)
                                    FROM playlists_media_files pmf
                                    WHERE pmf.playlist_id = ?),
                last_used = NOW()
            WHERE id = ?"""

        withTransaction {
            def params = [playlistId, mediaFileId]
            logger.debug(
                'Finding media file playlist position.\n\tSQL: {}\n\tPARAMS: {}',
                getPositionQuery, params)
            int position = sql.firstRow(getPositionQuery, params)[0]

            logger.debug(
                'Removing media file from playlist.\n\tSQL: {}\n\tPARAMS: {}',
                delQuery, params)
            sql.execute(delQuery, params)

            params = [playlistId, position]
            logger.debug('Reording playlist items.\n\tSQL: {}\n\tPARAMS: {}',
                reorderPlaylistQuery, params)
            sql.execute(reorderPlaylistQuery, params)

            params = [playlistId, playlistId]
            logger.debug('Updating playlist.\n\tSQL: {}\n\tPARAMS: {}',
                updatePlaylistQuery, params)
            sql.executeUpdate(updatePlaylistQuery, params)

            return getById(Playlist, playlistId) } }

    public Playlist removeAllFromPlaylist(int playlistId) {
        return withTransaction {
            def p = getById(Playlist, playlistId)
            if (!p) return null
            String query =
                'DELETE FROM playlists_media_files WHERE playlist_id = ?'
            def sqlParams = [playlistId]
            logger.debug('Clearing playlist.\n\tSQL: {}\n\tPARAMS: {}',
                query, sqlParams)
            sql.execute(query, sqlParams)
            p.mediaFileCount = 0
            p.lastUsed = new Timestamp(new Date().time)
            return update(p) } }

    public void removeEmptyPlaylists() {

        String query = """\
            DELETE FROM playlists WHERE id IN (
                SELECT DISTINCT p.id
                FROM playlists p LEFT OUTER JOIN playlists_media_files pmf ON
                    p.id = pmf.playlist_id
                WHERE pmf.playlist_id IS NULL)"""

        logger.debug('Deleting empty playlists.\n\tSQL: {}', query)
        sql.execute(query) }

    /// ### Tag-specific methods
    public Tag getTagById(int id) { return getById(Tag, id) }
    public Tag getTagByName(String name) { return getByName(Tag, name)[0] }
    public List<Tag> getTags() { return getAll(Tag, 'name ASC') }

    public List<Tag> getTagsWhere(Map params) {
        def query = new StringBuilder()
        def sqlParams = []

        query.append('SELECT DISTINCT t.* FROM tags t ')

        if (params.mediaFileId || params.artistId || params.albumId ||
            params.playlistId)
            query.append(' JOIN media_files_tags mft ON t.id = mft.tag_id')

        if (params.mediaFileId) {
            query.append(' AND mft.media_file_id = ? ')
            sqlParams << params.mediaFileId }

        if (params.playlistId) {
            query.append(' JOIN playlists_media_files pmf ON ')
                 .append(' mft.media_file_id = pmf.media_file_id AND ')
                 .append(' pmf.playlist_id = ? ')
            sqlParams << params.playlistId }

        if (params.artistId) {
            query.append(' JOIN artists_media_files armf ON ')
                 .append(' mft.media_file_id = armf.media_file_id AND ')
                 .append(' armf.artist_id = ? ')
            sqlParams << params.artistId }

        if (params.albumId) {
            query.append(' JOIN albums_media_files almf ON ')
                 .append(' mft.media_file_id = almf.media_file_id AND ')
                 .append(' almf.album_id = ? ')
            sqlParams << params.albumId }

        if (params.name) {
            query.append(' WHERE t.name = ? ')
            sqlParams << params.name }

        query = query.toString()
        logger.debug('Selecting tags.\n\tSQL: {}\n\tPARAMS: {}',
            query, sqlParams)
        return sql.rows(query, sqlParams).collect { recordToModel(Tag, it) } }

    public def tagMediaFiles(List<Integer> mediaFileIds, List<String> tagNames) {
        String insertQuery = 'INSERT INTO media_files_tags VALUES (?, ?)'
        String checkQuery = 'SELECT * FROM media_files_tags WHERE media_file_id = ? AND tag_id = ?'
        def params

        withTransaction {
            List<Tag> tags = tagNames.collect { name ->
                getTagByName(name) ?: create(new Tag(name: name)) }

            tags.each { tag ->
                params = [null, tag.id]
                mediaFileIds.each { mediaFileId ->
                    params[0] = mediaFileId
                    logger.debug('Checking to see if a media file is ' +
                        'already tagged.\n\tSQL: {}\n\tPARAMS: {}', params)
                    if (sql.rows(checkQuery, params).size() == 0) {
                        logger.debug(
                            'Tagging a media file.\n\tSQL: {}\n\tPARAMS: {}',
                            params)
                        sql.executeInsert(insertQuery, params) } } } } }

    public def untagMediaFiles(List<Integer> mediaFileIds,
    List<String> tagNames) {

        withTransaction {
            List<Tag> tags = tagNames.collect(this.&getTagByName).findAll()
            def tagPlaceHolders = tags.collect { '?' }.join(', ')
            def tagIds = tags.collect { it.id }

            String deleteQuery = """
                DELETE FROM media_files_tags
                WHERE tag_id IN ($tagPlaceHolders) AND media_file_id = ?"""

            mediaFileIds.each { mfId ->
                def params = tagIds + mfId
                logger.debug('Removing tags.\n\tSQL: {}\n\tPARAMS: {}',
                    deleteQuery, params)
                sql.execute(deleteQuery, params) } } }

    /// ### Utility functions
    public def withTransaction(Closure c) {
        try { sql.execute('BEGIN TRANSACTION'); return c() }
        finally { sql.execute('COMMIT') } }

    public static String nameToModel(String name) {
        def pts = name.toLowerCase().split('_')
        return pts.length == 1 ? pts[0] :
            pts[0] + pts[1..-1].collect { it.capitalize() }.join() }

    public static String nameFromModel(String name) {
        return UPPERCASE_PATTERN.matcher(name).
            replaceAll(/$1_$2/).toLowerCase() }

    public static String pluralize(String name) { return name + 's' }

    static def updateModel(def record, def model) {
        getInstanceFields(model.class).each { field ->
            field.set(model, record[nameFromModel(field.name)]) }
        return model }

    static def recordToModel(Class modelClass, def record) {
        if (record == null) return null

        def model = modelClass.newInstance()

        getInstanceFields(model.class).each { field ->
            field.set(model, record[nameFromModel(field.name)]) }

        return model }

    static def modelToRecord(def model) {
        if (model == null) return null

        def record = [:]

        getInstanceFields(model.class).each { field ->
            record[nameFromModel(field.name)] = field.get(model) }

        return record }

    static def getInstanceFields(Class modelClass) {
        return modelClass.fields.findAll { !Modifier.isStatic(it.modifiers) } }
}
